#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import json
import os
from unittest.mock import patch
from numpy import int8
from pandas import DataFrame as PandasDataFrame
from pandera import DataFrameSchema, Column, Check
from snowflake.snowpark import Session
from snowflake.snowpark import DataFrame as SnowparkDataFrame

from snowflake.snowpark_checkpoints.checkpoint import (
    _check_dataframe_schema_file,
    check_dataframe_schema,
    check_output_schema,
    check_input_schema,
)
from snowflake.snowpark import Session
from snowflake.snowpark.functions import lit

from snowflake.snowpark_checkpoints.utils.constant import (
    CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME,
    PASS_STATUS,
    SKIP_ALL,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
)


def test_input():
    checkpoint_name = "test_checkpoint"
    output__path = "test_output_path/unit/"
    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
        }
    )

    in_schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check(lambda x: 0 <= x <= 5, element_wise=True)),
            "COLUMN2": Column(float, Check(lambda x: x < -1.2)),
        }
    )

    @check_input_schema(in_schema, checkpoint_name, output_path=output__path)
    def preprocessor(dataframe: SnowparkDataFrame):
        dataframe = dataframe.withColumn(
            "column3", dataframe["COLUMN1"] + dataframe["COLUMN2"]
        )
        return dataframe

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)

    with patch(
        "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
    ) as mock_update_validation_result:
        preprocessor(sp_df)

    mock_update_validation_result.assert_called_once_with(
        checkpoint_name, PASS_STATUS, output__path
    )


def test_output():
    checkpoint_name = "test_checkpoint"
    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
        }
    )

    out_schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check(lambda x: 0 <= x <= 10, element_wise=True)),
            "COLUMN2": Column(float, Check(lambda x: x < -1.2)),
        }
    )

    @check_output_schema(out_schema, checkpoint_name)
    def preprocessor(dataframe: SnowparkDataFrame):
        return dataframe.with_column("COLUMN1", lit("Some bad data yo"))

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)

    with patch(
        "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
    ) as mock_update_validation_result:
        preprocessor(sp_df)

    mock_update_validation_result.assert_called_once_with(
        checkpoint_name, PASS_STATUS, None
    )


def test_df_check():
    checkpoint_name = "test_checkpoint"
    output_path = "test_output_path/unit/"
    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
        }
    )

    schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check(lambda x: 0 <= x <= 10)),
            "COLUMN2": Column(float, Check(lambda x: x < -1.2)),
        }
    )

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)

    with (
        patch(
            "snowflake.snowpark_checkpoints.checkpoint.is_checkpoint_enabled",
            return_value=True,
        ),
        patch(
            "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
        ) as mocked_update,
    ):
        check_dataframe_schema(sp_df, schema, checkpoint_name, output_path=output_path)

    mocked_update.assert_called_once_with(checkpoint_name, PASS_STATUS, output_path)


def test_df_check_from_file():
    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
            "COLUMN3": [True, False, True, False, True],
        }
    )

    schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check.between(0, 10)),
            "COLUMN2": Column(float, Check.between(-20.5, -1.0)),
        }
    )

    schema_data = {
        "pandera_schema": json.loads(schema.to_json()),
        "custom_data": {
            "columns": [
                {
                    "name": "COLUMN1",
                    "type": "integer",
                    "rows_count": 5,
                    "rows_not_null_count": 5,
                    "rows_null_count": 0,
                    "min": 0,
                    "max": 10,
                    "mean": 4.8,
                    "decimal_precision": 0,
                    "margin_error": 4.0693979898752,
                },
                {
                    "name": "COLUMN2",
                    "type": "float",
                    "rows_count": 5,
                    "rows_not_null_count": 5,
                    "rows_null_count": 0,
                    "min": -20.4,
                    "max": -1.3,
                    "mean": -7.22,
                    "decimal_precision": 1,
                    "margin_error": 7.3428604780426,
                },
            ],
        },
    }

    checkpoint_name = "test_checkpoint"

    current_directory_path = os.getcwd()

    output_directory_path = os.path.join(
        current_directory_path, SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME
    )

    if not os.path.exists(output_directory_path):
        os.makedirs(output_directory_path)

    checkpoint_file_name = CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME.format(
        checkpoint_name
    )

    checkpoint_file_path = os.path.join(output_directory_path, checkpoint_file_name)

    with open(checkpoint_file_path, "w") as output_file:
        output_file.write(json.dumps(schema_data))

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)

    with (
        patch(
            "snowflake.snowpark_checkpoints.checkpoint.is_checkpoint_enabled",
            return_value=True,
        ),
        patch(
            "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
        ) as mocked_update,
    ):
        _check_dataframe_schema_file(sp_df, checkpoint_name)

    mocked_update.assert_called_once_with(checkpoint_name, PASS_STATUS, None)


def test_df_check_custom_check():
    checkpoint_name = "test_checkpoint"

    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
        }
    )

    schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check(lambda x: 0 <= x <= 10, element_wise=True)),
            "COLUMN2": Column(float, Check(lambda x: x < -1.2)),
        }
    )

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)
    with (
        patch(
            "snowflake.snowpark_checkpoints.checkpoint.is_checkpoint_enabled",
            return_value=True,
        ),
        patch(
            "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
        ) as mocked_update,
    ):
        check_dataframe_schema(
            sp_df,
            schema,
            checkpoint_name,
            custom_checks={
                "COLUMN1": [
                    Check(lambda x: x.shape[0] == 5),
                    Check(lambda x: x.shape[1] == 2),
                ],
                "COLUMN2": [Check(lambda x: x.shape[0] == 5)],
            },
        )

    mocked_update.assert_called_once_with(checkpoint_name, PASS_STATUS, None)
    assert len(schema.columns["COLUMN1"].checks) == 3
    assert len(schema.columns["COLUMN2"].checks) == 2


def test_df_check_skip_check():
    checkpoint_name = "test_checkpoint"
    df = PandasDataFrame(
        {
            "COLUMN1": [1, 4, 0, 10, 9],
            "COLUMN2": [-1.3, -1.4, -2.9, -10.1, -20.4],
        }
    )

    schema = DataFrameSchema(
        {
            "COLUMN1": Column(int8, Check.between(0, 10, element_wise=True)),
            "COLUMN2": Column(
                float,
                [
                    Check.greater_than(-20.5),
                    Check.less_than(-1.0),
                    Check(lambda x: x < -1.2),
                ],
            ),
        }
    )

    session = Session.builder.getOrCreate()
    sp_df = session.create_dataframe(df)
    with (
        patch(
            "snowflake.snowpark_checkpoints.checkpoint.is_checkpoint_enabled",
            return_value=True,
        ),
        patch(
            "snowflake.snowpark_checkpoints.checkpoint._update_validation_result"
        ) as mocked_update,
    ):
        check_dataframe_schema(
            sp_df,
            schema,
            checkpoint_name,
            skip_checks={
                "COLUMN1": [SKIP_ALL],
                "COLUMN2": ["greater_than", "less_than"],
            },
        )

    mocked_update.assert_called_once_with(checkpoint_name, PASS_STATUS, None)
    assert len(schema.columns["COLUMN1"].checks) == 0
    assert len(schema.columns["COLUMN2"].checks) == 1



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from pandera import DataFrameSchema, Column, Check
from snowflake.snowpark import DataFrame as SnowparkDataFrame

from snowflake.snowpark_checkpoints.checkpoint import check_output_with
from snowflake.snowpark import Session

from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.spark_migration import check_with_spark
from snowflake.snowpark_checkpoints.spark_migration import SamplingStrategy
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession


session = Session.builder.getOrCreate()
spark_session = SparkSession.builder.getOrCreate()
job_context = SnowparkJobContext(session, spark_session, "demo-e2e-meaninful", True)


def my_spark_fn(df: SparkDataFrame):
    return df.filter(df.A < 2.0)


out_schema = DataFrameSchema(
    {
        "A": Column(float, Check(lambda x: 3 <= x <= 10, element_wise=True)),
        "B": Column(float, Check(lambda x: x < 5)),
    }
)


@check_with_spark(
    job_context=job_context,
    spark_function=my_spark_fn,
    sample=100,
    sampling_strategy=SamplingStrategy.RANDOM_SAMPLE,
)
@check_output_with(
    out_schema,
    sample=100,
    sampling_strategy=SamplingStrategy.RANDOM_SAMPLE,
    job_context=job_context,
)
def my_snowpark_fn(df: SnowparkDataFrame):
    return df.filter(df.a > 2.0)


df = job_context.snowpark_session.create_dataframe(
    [[1.1, 2.2], [3.3, 4.4]], schema=["a", "b"]
)

my_snowpark_fn(df)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import json
import tempfile
from unittest.mock import MagicMock, patch, mock_open
import unittest
from deepdiff import DeepDiff


class TelemetryManagerTest(unittest.TestCase):
    def test_get_telemetry_manager_no_setted(self):
        # Arrange
        from snowflake.snowpark_checkpoints.utils.telemetry import (
            TelemetryManager,
            get_telemetry_manager,
        )
        from snowflake.connector.telemetry import TelemetryClient

        _, mock_DIRS = mock_before_telemetry_import()
        session_mock = MagicMock()
        connection_mock = MagicMock(_telemetry=MagicMock(spec=TelemetryClient))
        session_mock.builder.getOrCreate.return_value = MagicMock(
            connection=connection_mock
        )
        type_mock = MagicMock()
        type_mock.__name__ = "TelemetryClient"

        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.Session", session_mock
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ):
            # Act
            result = get_telemetry_manager()
            result2 = get_telemetry_manager()

            # Assert
            assert isinstance(result, TelemetryManager)
            assert result is result2

    def test_get_metadata(self):
        # Arrange
        expected_os = "os"
        expected_python_version = "0.0.0"
        expected_unique_id = "1234"
        expected_snowpark_version = "0.0.0"

        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.platform", expected_os
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.python_version",
            return_value=expected_python_version,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWPARK_VERSION",
            (0, 0, 0),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry._get_unique_id",
            return_value=expected_unique_id,
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import _get_metadata

            # Act
            result = _get_metadata()

            # Assert
            assert result.get("os_version") == expected_os
            assert result.get("python_version") == expected_python_version
            assert result.get("snowpark_version") == expected_snowpark_version
            assert result.get("device_id") == expected_unique_id

    def test_get_unique_id(self):
        # Arrange
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.getnode", return_value=1234
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import _get_unique_id

            expected_id = (
                "03ac674216f3e15c761ee1a5e255f067953623c8b388b4459e13f978d7c846f4"
            )

            # Act
            result = _get_unique_id()

            # Assert
            assert result == expected_id

    def test_free_up_space(self):
        # Arrange
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry._get_folder_size",
            return_value=100,
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import _free_up_space

            folder_path, json_path = mock_folder_path(MagicMock(st_size=50, st_mtime=1))
            max_size = 50

            # Act
            _free_up_space(folder_path, max_size)

            # Assert
            folder_path.glob.assert_called_once_with("*.json")
            json_path.unlink.assert_called_once()

    def test_get_folder_size(self):
        # Arrange
        folder_path, json_path = mock_folder_path(MagicMock(st_size=50))
        from snowflake.snowpark_checkpoints.utils.telemetry import _get_folder_size

        expected_value = 100

        # Act
        result = _get_folder_size(folder_path)

        # Assert
        assert result == expected_value
        assert result == expected_value

    def test_generate_event(self):
        # Arrange
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_CONNECTOR_VERSION",
            "0.0.0",
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry._get_metadata",
            return_value={},
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.time_util.get_time_millis",
            return_value=0,
        ):

            mock_connection = MagicMock()
            mock_connection.application = "foo"
            from snowflake.snowpark_checkpoints.utils.telemetry import _generate_event

            # Act
            result = _generate_event("event_name", "event_type", {"testing": "boo"})

            # Assert
            assert result.get("message").get("type") == "event_type"
            assert result.get("message").get("event_name") == "event_name"
            assert result.get("message").get("driver_type") == "PythonConnector"
            assert result.get("message").get("source") == "snowpark-checkpoints"
            assert result.get("message").get("metadata") == {}
            assert result.get("message").get("data") == '{"testing": "boo"}'
            assert result.get("timestamp") == "0"
            assert result.get("message").get("driver_version") == "0.0.0"

    def test_telemetry_manager_upload_local_telemetry_success(self):
        # Arrange
        from pathlib import Path

        rest_mock, mock_DIRS = mock_before_telemetry_import()

        _, json_path = mock_folder_path(MagicMock(st_size=50))

        with (
            patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS",
                mock_DIRS,
            ),
            patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
                return_value=False,
            ),
            patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
                return_value=False,
            ),
            patch("builtins.open", mock_open(read_data='{"foo": "bar"}')),
            patch.object(Path, "glob", return_value=[json_path]),
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            # Act
            # Calls the upload_local_telemetry method in the __init__ method
            TelemetryManager(rest_mock)

            # Assert
            rest_mock.request.assert_called_with(
                "/telemetry/send",
                body={"logs": [{"foo": "bar"}]},
                method="post",
                client=None,
                timeout=5,
            )
            json_path.unlink.assert_called_once()

    def test_telemetry_manager_upload_local_telemetry_failed(self):
        # Arrange
        from pathlib import Path

        rest_mock, mock_DIRS = mock_before_telemetry_import(request_return=False)
        _, json_path = mock_folder_path(MagicMock(st_size=50))
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "builtins.open", mock_open(read_data='{"foo": "bar"}')
        ), patch.object(
            Path, "glob", return_value=[json_path]
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            # Act
            # Calls the upload_local_telemetry method in the __init__ method
            TelemetryManager(rest_mock)

            # Assert
            rest_mock.request.assert_called_with(
                "/telemetry/send",
                body={"logs": [{"foo": "bar"}]},
                method="post",
                client=None,
                timeout=5,
            )
            json_path.unlink.assert_not_called()

    def test_telemetry_manager_is_telemetry_enabled(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.getenv",
            return_value=None,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            # Act
            # Calls the is_telemetry_enabled method in the __init__ method
            telemetry = TelemetryManager(rest_mock)

            # Assert
            assert telemetry.sc_is_enabled == True
            TelemetryManager._sc_upload_local_telemetry.assert_called_once()

    def test_telemetry_manager_is_telemetry_disable(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.getenv",
            return_value=None,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            # Act
            # Calls the is_telemetry_enabled method in the __init__ method
            telemetry = TelemetryManager(None)

            # Assert
            assert telemetry.sc_is_enabled == False
            TelemetryManager._sc_upload_local_telemetry.assert_called_once()

    def test_telemetry_manager_is_telemetry_disable_env(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        rest_mock.telemetry_enabled = True
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.getenv",
            return_value="false",
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ):

            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            # Act
            # Calls the is_telemetry_enabled method in the __init__ method
            telemetry = TelemetryManager(rest_mock)

            # Assert
            assert telemetry.sc_is_enabled == False
            TelemetryManager._sc_upload_local_telemetry.assert_called_once()

    def test_telemetry_manager_write_telemetry(self):
        # Arrange
        mock_date = "00-00-0000_00-00-00"
        event = {"message": {"type": "test"}}
        batch = [event]
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_validate_folder_space",
            return_value=f"{batch[0]}",
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry.makedirs"
        ), patch(
            "builtins.open", mock_open()
        ), patch(
            "datetime.datetime",
            MagicMock(
                now=MagicMock(
                    return_value=MagicMock(strftime=MagicMock(return_value=mock_date))
                )
            ),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager
            from os import path

            telemetry = TelemetryManager(rest_mock)

            # Act
            telemetry._sc_write_telemetry(batch)

            # Assert
            TelemetryManager._sc_validate_folder_space.assert_called_once_with(batch[0])
            open.assert_called_once_with(
                telemetry.sc_folder_path / f"{mock_date}_telemetry_test.json", "w"
            )
            open().write.assert_called_once_with(f"{event}")

    def test_telemetry_manager_send_batch_no_telemetry(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)

            # Act
            result = telemetry._sc_send_batch([])

            # Assert
            assert result == False

    def test_telemetry_manager_send_batch_no_connection(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        rest_mock = None

        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_write_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)

            # Act
            result = telemetry._sc_send_batch([])

            # Assert
            TelemetryManager._sc_write_telemetry.assert_called_once_with([])
            assert result == False

    def test_telemetry_manager_send_batch_request_true(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)
            event = {"message": {"type": "test"}}
            batch = [event]

            # Act
            result = telemetry._sc_send_batch(batch)

            # Assert
            rest_mock.request.assert_called_with(
                "/telemetry/send",
                body={"logs": batch},
                method="post",
                client=None,
                timeout=5,
            )
            assert result == True

    def test_telemetry_manager_send_batch_request_false(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import(request_return=False)
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_write_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)
            event = {"message": {"type": "test"}}
            batch = [event]

            # Act
            result = telemetry._sc_send_batch(batch)

            # Assert
            rest_mock.request.assert_called_with(
                "/telemetry/send",
                body={"logs": batch},
                method="post",
                client=None,
                timeout=5,
            )
            TelemetryManager._sc_write_telemetry.assert_called_once_with(batch)
            assert result == False

    def test_telemetry_manager_sc_add_log_to_batch(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import(request_return=False)
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_send_batch",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)
            telemetry.sc_log_batch = []
            telemetry.sc_flush_size = 2
            event = {"foo": "boo"}

            # Act
            telemetry._sc_add_log_to_batch(event)

            # Assert
            telemetry._sc_send_batch.assert_not_called()
            assert len(telemetry.sc_log_batch) == 1
            assert telemetry.sc_log_batch[0] == event

    def test_telemetry_manager_sc_add_log_to_batch_and_clean(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import(request_return=False)
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_send_batch",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)
            telemetry.sc_log_batch = []
            telemetry.sc_flush_size = 1
            event = {"foo": "boo"}

            # Act
            telemetry._sc_add_log_to_batch(event)

            # Assert
            telemetry._sc_send_batch.assert_called_once_with([event])
            assert len(telemetry.sc_log_batch) == 0
            assert telemetry.sc_log_batch == []

    def test_telemetry_manager_log_telemetry(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_add_log_to_batch",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry._generate_event",
            return_value={"foo": "boo"},
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager
            from snowflake.snowpark_checkpoints.utils.telemetry import _generate_event

            telemetry = TelemetryManager(rest_mock)

            # Act
            result = telemetry._sc_log_telemetry(
                "event_name", "event_type", {"testing": "boo"}
            )

            # Assert
            TelemetryManager._sc_add_log_to_batch.assert_called_once_with(
                {"foo": "boo"}
            )
            _generate_event.assert_called_once_with(
                "event_name",
                "event_type",
                {"testing": "boo"},
            )
            assert result == {"foo": "boo"}

    def test_telemetry_manager_log_telemetry_disable(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_add_log_to_batch",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry._generate_event"
        ) as patch_generate_event:
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)

            # Act
            result = telemetry._sc_log_telemetry(
                "event_name", "event_type", {"testing": "boo"}
            )

            # Assert
            assert result.__eq__(None)
            telemetry._sc_add_log_to_batch.assert_not_called()
            patch_generate_event.assert_not_called()

    def test_telemetry_manager_log_info(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_log_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)

            # Act
            telemetry.sc_log_info("event_name", {"testing": "boo"})

            # Assert
            TelemetryManager._sc_log_telemetry.assert_called_once_with(
                "event_name", "info", {"testing": "boo"}
            )

    def test_telemetry_manager_log_error(self):
        # Arrange
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_log_telemetry",
            return_value=MagicMock(),
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ):
            from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

            telemetry = TelemetryManager(rest_mock)

            # Act
            telemetry.sc_log_error("event_name", {"testing": "boo"})

            # Assert
            TelemetryManager._sc_log_telemetry.assert_called_once_with(
                "event_name", "error", {"testing": "boo"}
            )

    def test_telemetry_manager_sc_is_telemetry_manager(self):
        # Arrange
        from snowflake.snowpark_checkpoints.utils.telemetry import TelemetryManager

        rest_mock, mock_DIRS = mock_before_telemetry_import()
        type_mock = MagicMock()

        with patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS", mock_DIRS
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
            return_value=True,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
            return_value=False,
        ), patch(
            "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
            return_value=MagicMock(),
        ):
            telemetry = TelemetryManager(rest_mock)

            # Act
            result = telemetry._sc_is_telemetry_manager()

            # Assert
            assert result

    def test_get_snowflake_schema_types(self):
        # Arrange
        from snowflake.snowpark_checkpoints.utils.telemetry import (
            get_snowflake_schema_types,
        )
        from snowflake.snowpark import dataframe as snowpark_dataframe
        from snowflake.snowpark.types import LongType, StructField

        snowpark_df = MagicMock(spec=snowpark_dataframe.DataFrame)
        snowpark_df.schema.fields = [StructField("foo", LongType(), True)]

        # Act

        result = get_snowflake_schema_types(snowpark_df)

        # Assert
        assert result == ["LongType()"]

    def test_get_spark_schema_types(self):
        # Arrange
        from snowflake.snowpark_checkpoints.utils.telemetry import (
            _get_spark_schema_types,
        )
        from pyspark.sql import dataframe as spark_dataframe
        from pyspark.sql.types import LongType, StructField

        spark_df = MagicMock(spec=spark_dataframe.DataFrame)
        spark_df.schema.fields = [StructField("foo", LongType(), True)]

        # Act

        result = _get_spark_schema_types(spark_df)

        # Assert
        assert result == ["LongType()"]

    def test_sc_write_telemetry_with_temp_folder(self):
        # Arrange
        from pathlib import Path
        from os import getcwd

        event = {"message": {"type": "test"}}
        batch = [event]
        rest_mock, mock_DIRS = mock_before_telemetry_import()
        mock_date = "00-00-0000-00-00-00"

        with tempfile.TemporaryDirectory(dir=getcwd()) as temp_dir:
            temp_path = Path(temp_dir)
            with patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.SNOWFLAKE_DIRS",
                mock_DIRS,
            ), patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_validate_folder_space",
                return_value=f"{json.dumps(event)}",
            ), patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_enabled",
                return_value=True,
            ), patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_is_telemetry_testing",
                return_value=False,
            ), patch(
                "snowflake.snowpark_checkpoints.utils.telemetry.TelemetryManager._sc_upload_local_telemetry",
                return_value=MagicMock(),
            ), patch(
                "datetime.datetime",
                MagicMock(
                    now=MagicMock(
                        return_value=MagicMock(
                            strftime=MagicMock(return_value=mock_date)
                        )
                    )
                ),
            ):
                from snowflake.snowpark_checkpoints.utils.telemetry import (
                    TelemetryManager,
                )

                telemetry = TelemetryManager(rest_mock)

                # Act
                telemetry.set_sc_output_path(temp_path)
                telemetry._sc_write_telemetry(batch)

                # Assert
                file_path = temp_path / f"{mock_date}_telemetry_test.json"
                self.assertTrue(file_path.exists(), "File was not created")
                with open(file_path) as file:
                    telemetry_output = json.load(file)
                TelemetryManager._sc_validate_folder_space.assert_called_once_with(
                    batch[0]
                )

                diff_telemetry = DeepDiff(
                    event,
                    telemetry_output,
                    ignore_order=True,
                )
                assert diff_telemetry == {}


def mock_folder_path(stat):
    folder_path = MagicMock()
    json_path = MagicMock()
    json_path.stat.return_value = stat
    json_path.is_file.return_value = True
    folder_path.glob.return_value = [json_path, json_path]

    return folder_path, json_path


def mock_before_telemetry_import(request_return=True):
    from pathlib import Path

    rest_mock = MagicMock()
    rest_mock.request.return_value = {"success": request_return}
    mock_DIRS = MagicMock()
    mock_DIRS.user_config_path = Path("folder/")
    return rest_mock, mock_DIRS



import os
from unittest.mock import mock_open, patch
from snowflake.snowpark_checkpoints.utils.constant import (
    PASS_STATUS,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    VALIDATION_RESULTS_JSON_FILE_NAME,
)
from snowflake.snowpark_checkpoints.validation_result_metadata import (
    ValidationResultsMetadata,
)
from snowflake.snowpark_checkpoints.validation_results import (
    ValidationResult,
    ValidationResults,
)


def test_load_with_valid_file():
    test_path = os.path.dirname(os.path.abspath(__file__))
    result_path = os.path.join(
        test_path,
        SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
        VALIDATION_RESULTS_JSON_FILE_NAME,
    )
    with open(result_path) as file:
        validation_result_json = file.read()
        mock_validation_results = ValidationResults.model_validate_json(
            validation_result_json
        )

    metadata = ValidationResultsMetadata(test_path)

    assert metadata.validation_results == mock_validation_results


def test_load_with_no_file():
    path = "mock_path"
    metadata = ValidationResultsMetadata(path)

    assert metadata.validation_results == ValidationResults(results=[])


def test_add_validation_result():
    path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(
        path,
        "checkpoint_validation_results",
        "non_existent_file.json",
    )

    metadata = ValidationResultsMetadata(path)

    new_validation_result = ValidationResult(
        timestamp="2021-01-01T00:00:00",
        file="file2",
        line_of_code=1,
        result=PASS_STATUS,
        checkpoint_name="checkpoint2",
    )

    metadata.add_validation_result(new_validation_result)

    assert metadata.validation_results == ValidationResults(
        results=[new_validation_result]
    )


def test_save_success():
    test_path = os.path.dirname(os.path.abspath(__file__))
    result_path = os.path.join(
        test_path,
        SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
        VALIDATION_RESULTS_JSON_FILE_NAME,
    )
    with open(result_path) as file:
        validation_result_json = file.read()
        mock_validation_results = ValidationResults.model_validate_json(
            validation_result_json
        )

    metadata = ValidationResultsMetadata(test_path)
    with patch("builtins.open", mock_open()) as mock_open_file:
        metadata.save()

    mock_open_file().write.assert_called_once_with(
        mock_validation_results.model_dump_json()
    )



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.spark_migration import check_with_spark
from snowflake.snowpark_checkpoints.spark_migration import SamplingStrategy
from pyspark.sql import DataFrame as SparkDataFrame
from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark import Session
from pyspark.sql import SparkSession
import pytest
import pandas as pd
import numpy as np


@pytest.fixture
def job_context():
    session = Session.builder.getOrCreate()
    spark_session = SparkSession.builder.getOrCreate()
    return SnowparkJobContext(session, spark_session, "test job", True)


def test_spark_checkpoint_scalar_passing(job_context):
    def my_spark_scalar_fn(df: SparkDataFrame):
        return df.count()

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_scalar_fn,
        checkpoint_name="test_spark_scalar_fn",
    )
    def my_snowpark_scalar_fn(df: SnowparkDataFrame):
        return df.count()

    df = job_context.snowpark_session.create_dataframe(
        [[1, 2], [3, 4]], schema=["a", "b"]
    )
    count = my_snowpark_scalar_fn(df)
    assert count == 2


def test_spark_checkpoint_scalar_fail(job_context):
    def my_spark_scalar_fn(df: SparkDataFrame):
        return df.count() + 1

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_scalar_fn,
        checkpoint_name="test_spark_scalar_fn",
    )
    def my_snowpark_scalar_fn(df: SnowparkDataFrame):
        return df.count()

    df = job_context.snowpark_session.create_dataframe(
        [[1, 2], [3, 4]], schema=["a", "b"]
    )
    try:
        my_snowpark_scalar_fn(df)
        assert (False, "Should have failed")
    except:
        pass


def test_spark_checkpoint_df_pass(job_context):
    def my_spark_fn(df: SparkDataFrame):
        return df.filter(df.A > 2)

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_fn,
        checkpoint_name="test_spark_scalar_fn",
    )
    def my_snowpark_fn(df: SnowparkDataFrame):
        return df.filter(df.a > 2)

    df = job_context.snowpark_session.create_dataframe(
        [[1, 2], [3, 4]], schema=["a", "b"]
    )
    my_snowpark_fn(df)


def test_spark_checkpoint_df_fail(job_context):
    def my_spark_fn(df: SparkDataFrame):
        return df.filter(df.A > 2)

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_fn,
        checkpoint_name="test_spark_scalar_fn",
    )
    def my_snowpark_fn(df: SnowparkDataFrame):
        return df.filter(df.a < 2)

    df = job_context.snowpark_session.create_dataframe(
        [[1, 2], [3, 4]], schema=["a", "b"]
    )
    try:
        my_snowpark_fn(df)
        assert (False, "Should have failed")
    except:
        pass


def test_spark_checkpoint_limit_sample(job_context):
    def my_spark_fn(df: SparkDataFrame):
        return df.filter(df.A < 50)

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_fn,
        checkpoint_name="test_spark_scalar_fn",
        sample_number=10,
        sampling_strategy=SamplingStrategy.LIMIT,
    )
    def my_snowpark_fn(df: SnowparkDataFrame):
        return df.filter(df.a < 50)

    data_df = pd.DataFrame(
        np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD")
    )
    df = job_context.snowpark_session.create_dataframe(data_df)
    my_snowpark_fn(df)


def test_spark_checkpoint_random_sample(job_context):
    def my_spark_fn(df: SparkDataFrame):
        return df.filter(df.A < 50)

    @check_with_spark(
        job_context=job_context,
        spark_function=my_spark_fn,
        checkpoint_name="test_spark_scalar_fn",
        sample_number=10,
        sampling_strategy=SamplingStrategy.RANDOM_SAMPLE,
    )
    def my_snowpark_fn(df: SnowparkDataFrame):
        return df.filter(df.a < 50)

    data_df = pd.DataFrame(
        np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD")
    )
    df = job_context.snowpark_session.create_dataframe(data_df)
    my_snowpark_fn(df)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import os
from unittest.mock import MagicMock, patch

from snowflake.snowpark_checkpoints.utils.constant import (
    SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR,
)


def test_is_checkpoint_import_error():
    with patch(
        "snowflake.snowpark_checkpoints_configuration.checkpoint_metadata.CheckpointMetadata",
        side_effect=ImportError("Mocked exception"),
    ):
        from snowflake.snowpark_checkpoints.utils.extra_config import (
            _get_metadata,
        )

        enabled, _ = _get_metadata()
        assert enabled == False


def test_is_checkpoint_enabled_default():
    with patch(
        "snowflake.snowpark_checkpoints.utils.extra_config._get_metadata",
        return_value=(False, None),
    ):
        from snowflake.snowpark_checkpoints.utils.extra_config import (
            is_checkpoint_enabled,
        )

        actual = is_checkpoint_enabled("demo_initial_creation_checkpoint")
        assert actual


def test_is_checkpoint_enabled_no_checkpoint_name():
    with patch(
        "snowflake.snowpark_checkpoints.utils.extra_config._get_metadata",
        return_value=(False, None),
    ):
        from snowflake.snowpark_checkpoints.utils.extra_config import (
            is_checkpoint_enabled,
        )

        actual = is_checkpoint_enabled()
        assert actual


def test_is_checkpoint_enabled_no_file():
    from snowflake.snowpark_checkpoints.utils.extra_config import is_checkpoint_enabled

    actual = is_checkpoint_enabled("demo_initial_creation_checkpoint")
    assert actual == True


def test_is_checkpoint_enabled_checkpoint_disabled():
    metadata = MagicMock()
    metadata.get_checkpoint.return_value = MagicMock(enabled=False)
    with patch(
        "snowflake.snowpark_checkpoints.utils.extra_config._get_metadata",
        return_value=(True, metadata),
    ):
        from snowflake.snowpark_checkpoints.utils.extra_config import (
            is_checkpoint_enabled,
        )

        actual = is_checkpoint_enabled("my-checkpoint")
        assert actual == False


def test_get_checkpoint_contract_file_path_env_var_set():
    os.environ[SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR] = "/mock/path"
    from snowflake.snowpark_checkpoints.utils.extra_config import (
        _get_checkpoint_contract_file_path,
    )

    assert _get_checkpoint_contract_file_path() == "/mock/path"


def test_get_checkpoint_contract_file_path_env_var_not_set():
    os.environ.pop(SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR)
    from snowflake.snowpark_checkpoints.utils.extra_config import (
        _get_checkpoint_contract_file_path,
    )

    assert _get_checkpoint_contract_file_path() == os.getcwd()



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import json
import os
from unittest.mock import ANY, call, patch, mock_open
from numpy import float64

from pytest import mark, raises
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.utils.constant import (
    BOOLEAN_TYPE,
    CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME,
    CHECKPOINT_TABLE_NAME_FORMAT,
    DATAFRAME_CUSTOM_DATA_KEY,
    DATAFRAME_EXECUTION_MODE,
    DEFAULT_KEY,
    EXCEPT_HASH_AGG_QUERY,
    FAIL_STATUS,
    FLOAT_TYPE,
    NAME_KEY,
    NULL_COUNT_KEY,
    OVERWRITE_MODE,
    PASS_STATUS,
    ROWS_COUNT_KEY,
    SCHEMA_EXECUTION_MODE,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    TYPE_KEY,
    CheckpointMode,
)
from pandera import Column, Check, DataFrameSchema
import pandas as pd
from unittest.mock import MagicMock
import numpy as np
from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _add_null_checks,
    _compare_data,
    _process_sampling,
    _update_validation_result,
    _is_valid_checkpoint_name,
    _validate_checkpoint_name,
)
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import SamplingStrategy
from pandera import DataFrameSchema, Column, Check

from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _generate_schema,
    _add_numeric_checks,
    MEAN_KEY,
    DECIMAL_PRECISION_KEY,
    _skip_checks_on_schema,
    SKIP_ALL,
    _add_boolean_checks,
    TRUE_COUNT_KEY,
    FALSE_COUNT_KEY,
    MARGIN_ERROR_KEY,
)
from snowflake.snowpark_checkpoints.validation_results import ValidationResult


def test_skip_specific_check():
    schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0), Check.less_than(10)]),
            "col2": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    skip_checks = {"col1": ["greater_than"]}
    _skip_checks_on_schema(schema, skip_checks)

    assert len(schema.columns["col1"].checks) == 1
    assert schema.columns["col1"].checks[0].name == "less_than"
    assert len(schema.columns["col2"].checks) == 1


def test_skip_checks_on_nonexistent_column():
    schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0), Check.less_than(10)]),
            "col2": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    skip_checks = {"col3": [SKIP_ALL]}
    _skip_checks_on_schema(schema, skip_checks)

    assert len(schema.columns["col1"].checks) == 2
    assert len(schema.columns["col2"].checks) == 1


def test_skip_no_checks():
    schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0), Check.less_than(10)]),
            "col2": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    skip_checks = {}
    _skip_checks_on_schema(schema, skip_checks)

    assert len(schema.columns["col1"].checks) == 2
    assert len(schema.columns["col2"].checks) == 1


def test_add_boolean_checks():
    schema = DataFrameSchema(
        {
            "col1": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    additional_check = {
        TRUE_COUNT_KEY: 2,
        FALSE_COUNT_KEY: 1,
        MARGIN_ERROR_KEY: 0,
        ROWS_COUNT_KEY: 3,
    }

    _add_boolean_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 3  # initial check + 2 added checks

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [True, True, False]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_boolean_checks_with_margin_error():
    schema = DataFrameSchema(
        {
            "col1": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    additional_check = {
        TRUE_COUNT_KEY: 2,
        FALSE_COUNT_KEY: 1,
        MARGIN_ERROR_KEY: 1,
        ROWS_COUNT_KEY: 3,
    }

    _add_boolean_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 3  # initial check + 2 added checks

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [True, True, False, False]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_boolean_checks_no_true_count():
    schema = DataFrameSchema(
        {
            "col1": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    additional_check = {FALSE_COUNT_KEY: 3, MARGIN_ERROR_KEY: 0, ROWS_COUNT_KEY: 3}

    _add_boolean_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 3  # initial check + 2 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [False, False, False]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_boolean_checks_no_false_count():
    schema = DataFrameSchema(
        {
            "col1": Column(bool, checks=[Check.isin([True, False])]),
        }
    )

    additional_check = {TRUE_COUNT_KEY: 3, MARGIN_ERROR_KEY: 0, ROWS_COUNT_KEY: 3}

    _add_boolean_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 3  # initial check + 2 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [True, True, True]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_numeric_checks():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {MEAN_KEY: 5.0, MARGIN_ERROR_KEY: 1.0}

    _add_numeric_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [4.5, 5.5, 5.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_numeric_checks_with_decimal_precision():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {
        MEAN_KEY: 5.0,
        MARGIN_ERROR_KEY: 1.0,
        DECIMAL_PRECISION_KEY: 2,
    }

    _add_numeric_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 3  # initial check + 2 added checks

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [4.50, 5.55, 5.00]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_numeric_checks_no_mean():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {MARGIN_ERROR_KEY: 1.0}

    _add_numeric_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [0.5, 1.5, 1.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_numeric_checks_no_margin_error():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {MEAN_KEY: 5.0}

    _add_numeric_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [5.0, 5.0, 5.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_numeric_checks_no_decimal_precision():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {MEAN_KEY: 5.0, MARGIN_ERROR_KEY: 1.0}

    _add_numeric_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [4.5, 5.5, 5.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_generate_schema_with_custom_data():
    schema = DataFrameSchema(
        {
            "col1": Column(float64, Check(lambda x: 0 <= x <= 10, element_wise=True)),
            "col2": Column(bool, Check.isin([True, False])),
        }
    )

    schema_data = {
        "pandera_schema": json.loads(schema.to_json()),
        "custom_data": {
            "columns": [
                {
                    "type": FLOAT_TYPE,
                    "name": "col1",
                    "min": 4.50,
                    "max": 5.55,
                    "mean": 5.01,
                    "decimal_precision": 2,
                    "margin_error": 0.42,
                },
                {
                    "type": BOOLEAN_TYPE,
                    "name": "col2",
                    "true_count": 2,
                    "false_count": 1,
                    "margin_error": 0.0,
                    "rows_count": 3,
                },
            ],
        },
    }

    checkpoint_name = "test_checkpoint"
    current_directory_path = os.getcwd()

    output_directory_path = os.path.join(
        current_directory_path, SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME
    )

    if not os.path.exists(output_directory_path):
        os.makedirs(output_directory_path)

    checkpoint_file_name = CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME.format(
        checkpoint_name
    )

    checkpoint_file_path = os.path.join(output_directory_path, checkpoint_file_name)

    with open(checkpoint_file_path, "w") as output_file:
        output_file.write(json.dumps(schema_data))

    schema = _generate_schema(checkpoint_name)

    assert "col1" in schema.columns
    assert "col2" in schema.columns
    assert len(schema.columns["col1"].checks) == 2  # 1 initial check + 1 added checks
    assert len(schema.columns["col2"].checks) == 3  # 1 initial check + 2 added checks

    df = pd.DataFrame({"col1": [4.50, 5.55, 5.00], "col2": [True, True, False]})
    schema.validate(df)


def test_generate_schema_no_pandera_schema_key():
    checkpoint_name = "test_checkpoint"
    schema_json = {
        DATAFRAME_CUSTOM_DATA_KEY: [
            {
                NAME_KEY: "col1",
                TYPE_KEY: "numeric",
                MEAN_KEY: 5.0,
                MARGIN_ERROR_KEY: 1.0,
            }
        ]
    }

    try:
        with patch("builtins.open", mock_open(read_data=json.dumps(schema_json))):
            schema = _generate_schema(checkpoint_name)
            assert False
    except:
        pass


def test_generate_schema_invalid_json():
    checkpoint_name = "test_checkpoint"
    invalid_json = "{invalid_json}"

    with patch("builtins.open", mock_open(read_data=invalid_json)):
        try:
            _generate_schema(checkpoint_name)
        except json.JSONDecodeError:
            assert True
        else:
            assert False


def test_process_sampling_with_default_params():
    # Mock Snowpark DataFrame
    df = MagicMock(spec=SnowparkDataFrame)
    df.write = MagicMock()
    df.write.mode = MagicMock(return_value=df.write)
    df.write.save_as_table = MagicMock()
    df.write = MagicMock()
    df.write.mode = MagicMock(return_value=df.write)
    df.write.save_as_table = MagicMock()
    df.count.return_value = 10

    # Mock Pandera schema
    pandera_schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0)]),
            "col2": Column(float, checks=[Check.less_than(10.0)]),
        }
    )

    # Mock job context
    job_context = MagicMock(spec=SnowparkJobContext)

    # Mock SamplingAdapter
    with patch(
        "snowflake.snowpark_checkpoints.utils.utils_checks.SamplingAdapter"
    ) as MockSamplingAdapter:
        mock_sampler = MockSamplingAdapter.return_value
        mock_sampler.get_sampled_pandas_args.return_value = [
            pd.DataFrame({"col1": [1, 2], "col2": [3.0, 4.0]})
        ]

        # Call the function
        adjusted_schema, sampled_df = _process_sampling(df, pandera_schema, job_context)

        # Assertions
        assert "COL1" in adjusted_schema.columns
        assert "COL2" in adjusted_schema.columns
        assert len(adjusted_schema.columns) == 2
        assert sampled_df.shape == (2, 2)
        assert np.all(sampled_df.index == 1)


def test_process_sampling_with_custom_params():
    # Mock Snowpark DataFrame
    df = MagicMock(spec=SnowparkDataFrame)
    df.count.return_value = pd.Series([10])

    # Mock Pandera schema
    pandera_schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0)]),
            "col2": Column(float, checks=[Check.less_than(10.0)]),
        }
    )

    # Mock job context
    job_context = MagicMock(spec=SnowparkJobContext)

    # Mock SamplingAdapter
    with patch(
        "snowflake.snowpark_checkpoints.utils.utils_checks.SamplingAdapter"
    ) as MockSamplingAdapter:
        mock_sampler = MockSamplingAdapter.return_value
        mock_sampler.get_sampled_pandas_args.return_value = [
            pd.DataFrame({"col1": [1, 2], "col2": [3.0, 4.0]})
        ]

        # Call the function with custom parameters
        adjusted_schema, sampled_df = _process_sampling(
            df,
            pandera_schema,
            job_context,
            sample_frac=0.2,
            sample_number=5,
            sampling_strategy=SamplingStrategy.RANDOM_SAMPLE,
        )

        # Assertions
        assert "COL1" in adjusted_schema.columns
        assert "COL2" in adjusted_schema.columns
        assert len(adjusted_schema.columns) == 2
        assert sampled_df.shape == (2, 2)
        assert np.all(sampled_df.index == 1)


def test_process_sampling_no_job_context():
    # Mock Snowpark DataFrame
    df = MagicMock(spec=SnowparkDataFrame)
    df.count.return_value = pd.Series([10])

    # Mock Pandera schema
    pandera_schema = DataFrameSchema(
        {
            "col1": Column(int, checks=[Check.greater_than(0)]),
            "col2": Column(float, checks=[Check.less_than(10.0)]),
        }
    )

    # Mock SamplingAdapter
    with patch(
        "snowflake.snowpark_checkpoints.utils.utils_checks.SamplingAdapter"
    ) as MockSamplingAdapter:
        mock_sampler = MockSamplingAdapter.return_value
        mock_sampler.get_sampled_pandas_args.return_value = [
            pd.DataFrame({"col1": [1, 2], "col2": [3.0, 4.0]})
        ]

        # Call the function without job context
        adjusted_schema, sampled_df = _process_sampling(df, pandera_schema)

        # Assertions
        assert "COL1" in adjusted_schema.columns
        assert "COL2" in adjusted_schema.columns
        assert len(adjusted_schema.columns) == 2
        assert sampled_df.shape == (2, 2)
        assert np.all(sampled_df.index == 1)


def test_compare_data_match():
    checkpoint_name = "test_checkpoint"
    new_checkpoint_name = CHECKPOINT_TABLE_NAME_FORMAT.format(checkpoint_name)

    # Mock Snowpark DataFrame
    df = MagicMock(spec=SnowparkDataFrame)

    # Mock job context
    job_context = MagicMock(spec=SnowparkJobContext)
    session = MagicMock()
    job_context.snowpark_session = session

    # Mock session.sql to return an empty DataFrame (indicating no mismatch)
    session.sql.return_value.count.return_value = 0

    checkpoint_name = "test_checkpoint"
    validation_status = PASS_STATUS
    output_path = "test_output_path/utils/"

    with (
        patch("os.getcwd", return_value="/mocked/path"),
        patch("os.path.exists", return_value=False),
        patch("builtins.open", mock_open()),
        patch("json.dump"),
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks._update_validation_result"
        ) as mock_update_validation_result,
    ):
        # Call the function
        _compare_data(df, job_context, checkpoint_name, output_path)

    # Assertions
    mock_update_validation_result.assert_called_once_with(
        checkpoint_name, validation_status, output_path
    )
    df.write.save_as_table.assert_called_once_with(
        table_name=new_checkpoint_name, mode=OVERWRITE_MODE
    )
    calls = [
        call(EXCEPT_HASH_AGG_QUERY, [checkpoint_name, new_checkpoint_name]),
        call().count(),
    ]
    session.sql.assert_has_calls(calls)
    job_context.mark_pass.assert_called_once_with(
        checkpoint_name, DATAFRAME_EXECUTION_MODE
    )
    job_context.mark_fail.assert_not_called()


def test_compare_data_mismatch():
    checkpoint_name = "test_checkpoint"
    new_checkpoint_name = CHECKPOINT_TABLE_NAME_FORMAT.format(checkpoint_name)

    # Mock Snowpark DataFrame
    df = MagicMock(spec=SnowparkDataFrame)

    # Mock job context
    job_context = MagicMock(spec=SnowparkJobContext)
    session = MagicMock()
    job_context.snowpark_session = session
    job_context.job_name = checkpoint_name

    # Mock session.sql to return a non-empty DataFrame (indicating a mismatch)
    session.sql.return_value.count.return_value = 1

    with (
        patch("os.getcwd", return_value="/mocked/path"),
        patch("os.path.exists", return_value=False),
        patch("builtins.open", mock_open()),
        patch("json.dump"),
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks._update_validation_result"
        ) as mock_update_validation_result,
    ):
        # Call the function and expect a SchemaValidationError
        with raises(
            SchemaValidationError,
            match=f"Data mismatch for checkpoint {checkpoint_name}",
        ):
            _compare_data(df, job_context, checkpoint_name)

    # Assertions
    mock_update_validation_result.assert_called_once_with(
        checkpoint_name, FAIL_STATUS, None
    )
    df.write.save_as_table.assert_called_once_with(
        table_name=new_checkpoint_name, mode=OVERWRITE_MODE
    )
    calls = [
        call(EXCEPT_HASH_AGG_QUERY, [checkpoint_name, new_checkpoint_name]),
        call().count(),
    ]
    session.sql.assert_has_calls(calls)
    job_context.mark_fail.assert_called()
    job_context.mark_pass.assert_not_called()


def test_update_validation_result_with_file():
    checkpoint_name = "test_checkpoint"
    validation_status = PASS_STATUS

    with (
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks.ValidationResultsMetadata"
        ) as MockValidationResultsMetadata,
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks.datetime"
        ) as mock_datetime,
    ):
        mock_datetime.now.return_value.isoformat.return_value = "2021-01-01T00:00:00"
        mock_pipeline_result_metadata = MockValidationResultsMetadata.return_value

        # Call the function
        _update_validation_result(checkpoint_name, validation_status)

        # Assertions
        MockValidationResultsMetadata.assert_called_once()

        mock_pipeline_result_metadata.add_validation_result.assert_called_once_with(
            ValidationResult(
                timestamp="2021-01-01T00:00:00",
                file=DEFAULT_KEY,
                line_of_code=-1,
                checkpoint_name=checkpoint_name,
                result=validation_status,
            )
        )
        mock_pipeline_result_metadata.save.assert_called_once()


def test_update_validation_result_without_file():
    checkpoint_name = "test_checkpoint"
    validation_status = PASS_STATUS

    with (
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks.ValidationResultsMetadata"
        ) as MockValidationResultsMetadata,
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks.inspect.stack",
            return_value=[
                MagicMock(
                    filename="test_file.py",
                    code_context=["_check_dataframe_schema_file"],
                ),
                MagicMock(
                    filename="test_file.py",
                    code_context=["validate_dataframe_checkpoint"],
                    lineno=1,
                ),
            ],
        ),
        patch(
            "snowflake.snowpark_checkpoints.utils.utils_checks.datetime"
        ) as mock_datetime,
    ):
        mock_datetime.now.return_value.isoformat.return_value = "2021-01-01T00:00:00"
        mock_pipeline_result_metadata = MockValidationResultsMetadata.return_value

        # Call the function
        _update_validation_result(checkpoint_name, validation_status)

        # Assertions
        MockValidationResultsMetadata.assert_called_once()
        mock_pipeline_result_metadata.add_validation_result.assert_called_once_with(
            ValidationResult(
                timestamp="2021-01-01T00:00:00",
                file="test_file.py",
                line_of_code=1,
                checkpoint_name=checkpoint_name,
                result=validation_status,
            )
        )
        mock_pipeline_result_metadata.save.assert_called_once()


@mark.parametrize("name", ["checkpoint1", "Checkpoint_2", "CHECKPOINT_3"])
def test_is_valid_checkpoint_name_valid(name: str):
    assert _is_valid_checkpoint_name(name)


@mark.parametrize(
    "name",
    ["checkpoint-1", "Checkpoint 2", "CHECKPOINT@3", "checkpoint!", "123checkpoint"],
)
def test_is_valid_checkpoint_name_invalid(name: str):
    assert _is_valid_checkpoint_name(name) is False


@mark.parametrize(
    "name",
    [
        "checkpoint-1",
        "Checkpoint 2",
        "CHECKPOINT@3",
        "checkpoint!",
        "123checkpoint",
    ],
)
def test_validate_checkpoint_name_invalid(name: str):
    with raises(
        ValueError,
        match=f"Invalid checkpoint name: {name}. Checkpoint names must only contain alphanumeric characters and underscores.",
    ):
        _validate_checkpoint_name(name)


def test_add_null_checks():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)], nullable=True),
        }
    )

    additional_check = {
        NULL_COUNT_KEY: 2,
        ROWS_COUNT_KEY: 5,
        MARGIN_ERROR_KEY: 0,
    }

    _add_null_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [1.0, None, 2.0, None, 3.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_null_checks_with_margin_error():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)], nullable=True),
        }
    )

    additional_check = {
        NULL_COUNT_KEY: 2,
        ROWS_COUNT_KEY: 5,
        MARGIN_ERROR_KEY: 1,
    }

    _add_null_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [1.0, None, 2.0, None, None]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_null_checks_no_null_count():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)]),
        }
    )

    additional_check = {ROWS_COUNT_KEY: 5, MARGIN_ERROR_KEY: 0}

    _add_null_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [1.0, 2.0, 3.0, 4.0, 5.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)


def test_add_null_checks_no_margin_error():
    schema = DataFrameSchema(
        {
            "col1": Column(float, checks=[Check.greater_than(0)], nullable=True),
        }
    )

    additional_check = {NULL_COUNT_KEY: 2, ROWS_COUNT_KEY: 5}

    _add_null_checks(schema, "col1", additional_check)

    assert len(schema.columns["col1"].checks) == 2  # initial check + 1 added check

    # Create a DataFrame to test the checks
    df = pd.DataFrame({"col1": [1.0, None, 2.0, None, 3.0]})

    # Validate the DataFrame against the schema
    schema.validate(df)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import pandas as pd
from pandas.testing import assert_frame_equal
from pyspark.sql import SparkSession
from pytest import fixture
from snowflake.snowpark import Session

from snowflake.snowpark_checkpoints.spark_migration import compare_spark_snowpark_dfs


@fixture
def spark_session():
    return SparkSession.builder.getOrCreate()


@fixture
def snowpark_session():
    return Session.builder.config("local_testing", True).getOrCreate()


@fixture
def data():
    return [("Raul", 21), ("John", 34), ("Rose", 50)]


def test_compare_spark_snowpark_dfs_same_columns(spark_session, snowpark_session, data):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    snowpark_df = snowpark_session.create_dataframe(data, schema=["name", "age"])
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert cmp.empty


def test_compare_spark_snowpark_dfs_different_column_names(
    spark_session, snowpark_session, data
):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    snowpark_df = snowpark_session.create_dataframe(data, schema=["full_name", "age"])
    expected_data = [
        ("Raul", None, "spark_only"),
        ("John", None, "spark_only"),
        ("Rose", None, "spark_only"),
        (None, "Raul", "snowpark_only"),
        (None, "John", "snowpark_only"),
        (None, "Rose", "snowpark_only"),
    ]

    expected_cmp = pd.DataFrame.from_records(
        expected_data, columns=["NAME", "FULL_NAME", "_merge"]
    )
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(expected_cmp, cmp, check_dtype=False, check_categorical=False)


def test_compare_spark_snowpark_dfs_column_left_mismatch(
    spark_session, snowpark_session, data
):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    spark_df = spark_df.drop("age")
    snowpark_df = snowpark_session.create_dataframe(data, schema=["name", "age"])

    expected_cmp = pd.DataFrame.from_records(data, columns=["NAME", "AGE"])
    expected_cmp = expected_cmp.drop(columns="NAME")
    expected_cmp["_merge"] = pd.Categorical(["snowpark_only"] * 3)
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(expected_cmp, cmp, check_dtype=False, check_categorical=False)


def test_compare_spark_snowpark_dfs_column_right_mismatch(
    spark_session, snowpark_session, data
):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    snowpark_df = snowpark_session.create_dataframe(data, schema=["name", "age"])
    snowpark_df = snowpark_df.drop("age")

    expected_cmp = pd.DataFrame.from_records(data, columns=["NAME", "AGE"])
    expected_cmp = expected_cmp.drop(columns="NAME")
    expected_cmp["_merge"] = pd.Categorical(["spark_only"] * 3)
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(expected_cmp, cmp, check_dtype=False, check_categorical=False)


def test_compare_spark_snowpark_dfs_right_more_rows(
    spark_session, snowpark_session, data
):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    spark_df = spark_df.filter(spark_df.age < 50)
    snowpark_df = snowpark_session.create_dataframe(data, schema=["name", "age"])

    expected_cmp = pd.DataFrame.from_records(data, columns=["NAME", "AGE"])
    expected_cmp = expected_cmp[expected_cmp["AGE"] >= 50]
    expected_cmp["_merge"] = pd.Categorical(["snowpark_only"])
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(expected_cmp, cmp, check_dtype=False, check_categorical=False)


def test_compare_spark_snowpark_dfs_left_more_rows(
    spark_session, snowpark_session, data
):
    spark_df = spark_session.createDataFrame(data, schema="name string, age integer")
    snowpark_df = snowpark_session.create_dataframe(data, schema=["name", "age"])
    snowpark_df = snowpark_df.filter(snowpark_df.age < 50)

    expected_cmp = pd.DataFrame.from_records(data, columns=["NAME", "AGE"])
    expected_cmp = expected_cmp[expected_cmp["AGE"] >= 50]
    expected_cmp["_merge"] = pd.Categorical(["spark_only"])
    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(expected_cmp, cmp, check_dtype=False, check_categorical=False)


def test_compare_spark_snowpark_dfs_different_rows(spark_session, snowpark_session):
    spark_data = [("John", 34), ("Raul", 21)]
    snowpark_data = [("Rose", 50)]
    spark_df = spark_session.createDataFrame(
        spark_data, schema="name string, age integer"
    )
    spark_df = spark_df.filter(spark_df.age < 50)
    snowpark_df = snowpark_session.create_dataframe(
        snowpark_data, schema=["name", "age"]
    )
    snowpark_df = snowpark_df.filter(snowpark_df.age >= 50)

    original_df = pd.DataFrame.from_records(
        spark_data + snowpark_data, columns=["NAME", "AGE"]
    )
    snowpark_cmp = original_df[original_df["AGE"] >= 50]
    snowpark_cmp["_merge"] = pd.Categorical(["snowpark_only"])
    spark_cmp = original_df[original_df["AGE"] < 50]
    spark_cmp["_merge"] = pd.Categorical(["spark_only"] * 2)

    expected_df = pd.concat([spark_cmp, snowpark_cmp], ignore_index=True)

    cmp = compare_spark_snowpark_dfs(spark_df, snowpark_df)
    assert_frame_equal(
        expected_df,
        cmp,
        check_dtype=False,
        check_categorical=False,
    )



from datetime import datetime
from unittest.mock import MagicMock, patch

from pyspark.sql import SparkSession
from pytest import fixture, raises
from snowflake.snowpark import Session
from snowflake.snowpark.types import (
    BooleanType,
    DateType,
    DoubleType,
    LongType,
    StringType,
    StructField,
    StructType,
)

from snowflake.snowpark_checkpoints.checkpoint import validate_dataframe_checkpoint
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.utils.constant import (
    DATAFRAME_EXECUTION_MODE,
    CheckpointMode,
    FAIL_STATUS,
    PASS_STATUS,
)


@fixture
def job_context():
    session = Session.builder.getOrCreate()
    job_context = SnowparkJobContext(
        session, SparkSession.builder.getOrCreate(), "real_demo", True
    )
    return job_context


@fixture
def spark_schema():
    import pyspark.sql.types as t

    return t.StructType(
        [
            t.StructField("BYTE", t.ByteType(), True),
            t.StructField("SHORT", t.ShortType(), True),
            t.StructField("INTEGER", t.IntegerType(), True),
            t.StructField("LONG", t.LongType(), True),
            t.StructField("FLOAT", t.FloatType(), True),
            t.StructField("DOUBLE", t.DoubleType(), True),
            # StructField("decimal", DecimalType(10, 3), True),
            t.StructField("STRING", t.StringType(), True),
            # StructField("binary", BinaryType(), True),
            t.StructField("BOOLEAN", t.BooleanType(), True),
            t.StructField("DATE", t.DateType(), True),
            # StructField("timestamp", TimestampType(), True),
            # StructField("timestamp_ntz", TimestampType(), True),
        ]
    )


@fixture
def snowpark_schema():
    return StructType(
        [
            StructField("BYTE", LongType(), True),
            StructField("SHORT", LongType(), True),
            StructField("INTEGER", LongType(), True),
            StructField("LONG", LongType(), True),
            StructField("FLOAT", DoubleType(), True),
            StructField("DOUBLE", DoubleType(), True),
            # StructField("decimal", DecimalType(10, 3), True),
            StructField("STRING", StringType(), True),
            # StructField("binary", BinaryType(), True),
            StructField("BOOLEAN", BooleanType(), True),
            StructField("DATE", DateType(), True),
            # StructField("timestamp", TimestampType(), True),
            # StructField("timestamp_ntz", TimestampType(), True),
        ]
    )


@fixture
def data():
    date_format = "%Y-%m-%d"
    timestamp_format = "%Y-%m-%d %H:%M:%S"
    timestamp_ntz_format = "%Y-%m-%d %H:%M:%S"

    return [
        [
            3,
            789,
            13579,
            1231231231,
            7.8,
            2.345678,
            # Decimal(7.891),
            "red",
            # b"info",
            True,
            datetime.strptime("2023-03-01", date_format),
            # datetime.strptime("2023-03-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-03-01 12:00:00", timestamp_ntz_format),
        ],
        [
            4,
            101,
            24680,
            3213213210,
            0.12,
            3.456789,
            # Decimal(0.123),
            "red",
            # b"test",
            False,
            datetime.strptime("2023-04-01", date_format),
            # datetime.strptime("2023-04-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-04-01 12:00:00", timestamp_ntz_format),
        ],
        [
            5,
            202,
            36912,
            4564564560,
            3.45,
            4.567890,
            # Decimal(3.456),
            "red",
            # b"example2",
            True,
            datetime.strptime("2023-05-01", date_format),
            # datetime.strptime("2023-05-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-05-01 12:00:00", timestamp_ntz_format),
        ],
        [
            6,
            303,
            48123,
            7897897890,
            6.78,
            5.678901,
            # Decimal(6.789),
            "red",
            # b"sample2",
            False,
            datetime.strptime("2023-06-01", date_format),
            # datetime.strptime("2023-06-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-06-01 12:00:00", timestamp_ntz_format),
        ],
        [
            7,
            404,
            59234,
            9879879870,
            9.01,
            6.789012,
            # Decimal(9.012),
            "red",
            # b"data2",
            True,
            datetime.strptime("2023-07-01", date_format),
            # datetime.strptime("2023-07-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-07-01 12:00:00", timestamp_ntz_format),
        ],
        [
            8,
            505,
            70345,
            1231231234,
            1.23,
            7.890123,
            # Decimal(1.234),
            "blue",
            # b"test2",
            False,
            datetime.strptime("2023-08-01", date_format),
            # datetime.strptime("2023-08-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-08-01 12:00:00", timestamp_ntz_format),
        ],
        [
            9,
            606,
            81456,
            3213213214,
            4.56,
            8.901234,
            # Decimal(4.567),
            "blue",
            # b"example3",
            True,
            datetime.strptime("2023-09-01", date_format),
            # datetime.strptime("2023-09-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-09-01 12:00:00", timestamp_ntz_format),
        ],
        [
            10,
            707,
            92567,
            4564564564,
            7.8,
            9.012345,
            # Decimal(7.892),
            "blue",
            # b"sample3",
            False,
            datetime.strptime("2023-10-01", date_format),
            # datetime.strptime("2023-10-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-10-01 12:00:00", timestamp_ntz_format),
        ],
        [
            11,
            808,
            103678,
            7897897894,
            0.12,
            0.123456,
            # Decimal(0.123),
            "green",
            # b"data3",
            True,
            datetime.strptime("2023-11-01", date_format),
            # datetime.strptime("2023-11-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-11-01 12:00:00", timestamp_ntz_format),
        ],
        [
            12,
            909,
            114789,
            9879879874,
            3.45,
            1.234567,
            # Decimal(3.456),
            "green",
            # b"test3",
            False,
            datetime.strptime("2023-12-01", date_format),
            # datetime.strptime("2023-12-01 12:00:00", timestamp_format),
            # datetime.strptime("2023-12-01 12:00:00", timestamp_ntz_format),
        ],
    ]


def test_df_mode_dataframe(job_context, snowpark_schema, data):
    stage_name = "test_df_mode_dataframe"
    checkpoint_name = "test_mode_dataframe_checkpoint"
    df = job_context.snowpark_session.create_dataframe(data, snowpark_schema)
    df.write.save_as_table(checkpoint_name, mode="overwrite")

    mocked_session = MagicMock()
    job_context.mark_pass = mocked_session

    with patch(
        "snowflake.snowpark_checkpoints.utils.utils_checks._update_validation_result"
    ) as mocked_update:
        validate_dataframe_checkpoint(
            df,
            checkpoint_name,
            job_context=job_context,
            mode=CheckpointMode.DATAFRAME,
        )

    mocked_update.assert_called_once_with(checkpoint_name, PASS_STATUS, None)
    mocked_session.assert_called_once_with(checkpoint_name, DATAFRAME_EXECUTION_MODE)


def test_df_mode_dataframe_mismatch(job_context, snowpark_schema, data):
    checkpoint_name = "test_mode_dataframe_checkpoint_fail"

    data_copy = data.copy()
    df = job_context.snowpark_session.create_dataframe(data_copy, snowpark_schema)
    df.write.save_as_table(checkpoint_name, mode="overwrite")

    data.pop()
    df_spark = job_context.snowpark_session.create_dataframe(data, snowpark_schema)

    with patch(
        "snowflake.snowpark_checkpoints.utils.utils_checks._update_validation_result"
    ) as mocked_update:
        with raises(
            SchemaValidationError,
            match=f"Data mismatch for checkpoint {checkpoint_name}",
        ):
            validate_dataframe_checkpoint(
                df_spark,
                checkpoint_name,
                job_context=job_context,
                mode=CheckpointMode.DATAFRAME,
            )

    mocked_update.assert_called_once_with(checkpoint_name, FAIL_STATUS, None)


def test_df_mode_dataframe_job_none(job_context, snowpark_schema, data):
    checkpoint_name = "test_mode_dataframe_checkpoint_fail"
    df_spark = job_context.snowpark_session.create_dataframe(data, snowpark_schema)

    with raises(
        ValueError, match="Connectionless mode is not supported for Parquet validation"
    ):
        validate_dataframe_checkpoint(
            df_spark,
            checkpoint_name,
            job_context=None,
            mode=CheckpointMode.DATAFRAME,
        )


def test_df_mode_dataframe_invalid_mode(job_context, snowpark_schema, data):
    checkpoint_name = "test_mode_dataframe_checkpoint_fail"
    df_spark = job_context.snowpark_session.create_dataframe(data, snowpark_schema)

    with raises(
        ValueError,
        match="""Invalid validation mode.
            Please use for schema validation use a 1 or for a full data validation use a 2 for schema validation.""",
    ):
        validate_dataframe_checkpoint(
            df_spark,
            checkpoint_name,
            job_context=job_context,
            mode="invalid",
        )
